package org.freedesktop.dbus;

import java.lang.reflect.Array;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.util.List;

class RemoteInvocationHandler implements InvocationHandler
{

   public static Object executeRemoteMethod(RemoteObject ro, Method m, DBusConnection conn, Class iface, Object[] args) throws DBusExecutionException
   {
      Class[] ts = m.getParameterTypes();
      try {
         args = DBusConnection.convertParameters(args, ts);
      } catch (Exception e) {
         e.printStackTrace();
         throw new DBusExecutionException(e.getMessage());
      }
      MethodCall call = new MethodCall(ro.service, ro.objectpath, ro.iface, m.getName(), args);
      synchronized (conn.outgoing) {
         conn.outgoing.add(call);
      }

      Class c = m.getReturnType();

      DBusMessage reply = call.getReply();
               
      Object[] rp = reply.getParameters();

      if (reply instanceof DBusErrorMessage)
         ((DBusErrorMessage) reply).throwException();
         
      if (null == rp) { 
         if(null == c || Void.TYPE.equals(c)) return null;
         else throw new DBusExecutionException("Wrong return type (expected void)");
      }

      switch (rp.length) {
         case 0:
            if (null == c || Void.TYPE.equals(c))
               return null;
            else throw new DBusExecutionException("Wrong return type (expected void)");
         case 1:

            try { 
               rp = DBusConnection.deSerialiseParameters(rp, 
                     new Class[] { m.getReturnType() });
            }
            catch (Exception e) { 
               throw new DBusExecutionException("Wrong return type (failed to de-serialise correct types: "+e.getMessage()+")");
            }

            return rp[0];
         default:

            return rp;
      }
   }

   DBusConnection conn;
   RemoteObject remote;
   Class iface;
   public RemoteInvocationHandler(DBusConnection conn, RemoteObject remote, Class iface)
   {
      this.iface = iface;
      this.remote = remote;
      this.conn = conn;
   }
   public Object invoke(Object proxy, Method method, Object[] args) throws Throwable
   {
      if (method.getName().equals("isRemote")) return Boolean.TRUE;
      else if (method.getName().equals("clone")) return null;
      else if (method.getName().equals("equals")) {
         try {
            if (1 == args.length) 
               return new Boolean(remote.equals(((RemoteInvocationHandler) Proxy.getInvocationHandler(args[0])).remote));
         } catch (IllegalArgumentException IAe) {
            return Boolean.FALSE;
         }
      }
      else if (method.getName().equals("finalize")) return null;
      else if (method.getName().equals("getClass")) return iface;
      else if (method.getName().equals("hashCode")) return new Integer(remote.hashCode());
      else if (method.getName().equals("notify")) {
         remote.notify();
         return null;
      } else if (method.getName().equals("notifyAll")) {
         remote.notifyAll();
         return null;
      } else if (method.getName().equals("wait"))  {
         if (0 == args.length) remote.wait();
         else if (1 == args.length 
               && args[0] instanceof Long) remote.wait(((Long) args[0]).longValue());
         else if (2 == args.length 
               && args[0] instanceof Long
               && args[1] instanceof Integer) 
            remote.wait(((Long) args[0]).longValue(), ((Integer) args[1]).intValue());
         if (args.length <= 2)
            return null;
      }
      else if (method.getName().equals("toString"))
         return remote.service+":"+remote.objectpath+":"+remote.iface;

      return executeRemoteMethod(remote, method, conn, iface, args);
   }
}

