package org.freedesktop.dbus.test;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import org.freedesktop.dbus.DBusConnection;
import org.freedesktop.dbus.DBusException;
import org.freedesktop.dbus.DBusExecutionException;
import org.freedesktop.dbus.DBusInterface;
import org.freedesktop.dbus.DBusSigHandler;
import org.freedesktop.dbus.DBusSignal;
import org.freedesktop.dbus.UInt16;
import org.freedesktop.dbus.UInt32;
import org.freedesktop.dbus.UInt64;
import org.freedesktop.dbus.Variant;

import org.freedesktop.DBus.Error.UnknownObject;
import org.freedesktop.DBus.Error.ServiceUnknown;
import org.freedesktop.DBus.Peer;
import org.freedesktop.DBus.Introspectable;

class testclass implements TestRemoteInterface, TestRemoteInterface2, TestSignalInterface
{
   public void dostuff(Variant foo)
   {
      System.out.println("Doing Stuff "+foo);
      if (!(foo instanceof Variant) ||
            !(foo.getValue() instanceof Boolean) ||
            ((Boolean) foo.getValue()).booleanValue() != true)
         test.fail("dostuff received the wrong argument");
   }
   /** Local classes MUST implement this to return false */
   public boolean isRemote() { return false; }
   /** The method we are exporting to the Bus. */
   public Integer[] sampleArray(String[] ss, Integer[] is, long[] ls)
   {
      System.out.println("Got an array:");
      for (int i = 0; i < ss.length; i++)
         System.out.println("--"+ss[i]);
      if (ss.length != 5 ||
            !"hi".equals(ss[0]) ||
            !"hello".equals(ss[1]) ||
            !"hej".equals(ss[2]) ||
            !"hey".equals(ss[3]) ||
            !"aloha".equals(ss[4]))
         test.fail("sampleArray, String array contents incorrect");
      System.out.println("Got an array:");
      for (int i = 0; i < is.length; i++)
         System.out.println("--"+is[i]);
      if (is.length != 4 ||
            is[0].intValue() != 1 ||
            is[1].intValue() != 5 ||
            is[2].intValue() != 7 ||
            is[3].intValue() != 9)
         test.fail("sampleArray, Integer array contents incorrect");
      System.out.println("Got an array:");
      for (int i = 0; i < ls.length; i++)
         System.out.println("--"+ls[i]);
      if (ls.length != 4 ||
            ls[0] != 2 ||
            ls[1] != 6 ||
            ls[2] != 8 ||
            ls[3] != 12)
         test.fail("sampleArray, Integer array contents incorrect");
      return new Integer[] { new Integer(-1), new Integer(-5), new Integer(-7), new Integer(-12), new Integer(-18) };
   }
   public String getName()
   {
      return "This Is A Name!!";
   }
   public short[] eatMyShorts()
   {
      return new short[] { 1, 2, 3 };
   }
   public void throwme() throws TestException
   {
      throw new TestException("test");
   }
}

/**
 * This is a test program which sends and recieves a signal, implements, exports and calls a remote method.
 */
public class test implements DBusSigHandler
{
   public static void fail(String message)
   {
      System.err.println("Test Failed: "+message);
      if (null != conn) conn.disconnect();
      System.exit(1);
   }
   /** Handling a signal */
   public void handle(DBusSignal s)
   {
      System.out.println("SignalHandler Running");

      /** Known signals which we have a class for will be instantiated to that class. */
      if (s instanceof TestSignalInterface.TestSignal) {
         System.out.print("Got a test signal with Parameters: ");
         /** Use the test signal interface to get the parameters. */
         TestSignalInterface.TestSignal t = (TestSignalInterface.TestSignal) s;
         System.out.println("string("+t.value+") int("+t.number+")");
         if (!"Bar".equals(t.value) || !(new UInt32(42)).equals(t.number))
            fail("Incorrect TestSignal parameters");
      } 
      /** Known signals which we have a class for will be instantiated to that class. */
      else if (s instanceof TestSignalInterface.TestArraySignal) {
         System.out.println("Got a test array signal with Parameters: ");
         /** Use the test signal interface to get the parameters. */
         TestSignalInterface.TestArraySignal t = (TestSignalInterface.TestArraySignal) s;
         for (int i = 0; i < t.v.length; i++)
            System.out.println("--"+t.v[i]);
         if (t.v.length != 5 ||
             !"hi".equals(t.v[0]) ||
             !"hello".equals(t.v[1]) ||
             !"hej".equals(t.v[2]) ||
             !"hey".equals(t.v[3]) ||
             !"aloha".equals(t.v[4]))
            fail("Incorrect TestArraySignal parameters");
      } 
      /** Other signals. (should never happen, we are registered for only TestSignals) */
      else {
         System.out.println("Got an unknown signal: "+s);
         /** Use the generic signal interface to get the details. */
         System.out.println("Class: "+s.getClass());
         System.out.println("Type: "+s.getType());
         System.out.println("Name: "+s.getName());
         System.out.print("Parameters:");
         for (int i = 0; i < s.getParameters().length; i++)
            System.out.print(" "+s.getParameters()[i]);
         System.out.println();
         fail("Incorrect Signal Type");
      }
   }
   static DBusConnection conn = null;
   public static void main(String[] args) 
   { try {
      System.out.println("Creating Connection");
      conn = DBusConnection.getConnection(DBusConnection.SESSION);
      
      System.out.println("Registering Name");
      conn.registerService("foo.bar.Test");
      
      System.out.print("Listening for signals...");
      try {
         /** This registers an instance of the test class as the signal handler for the TestSignal class. */
         test t = new test();
         conn.addSigHandler(TestSignalInterface.TestSignal.class, t);
         conn.addSigHandler(TestSignalInterface.TestArraySignal.class, t);
         System.out.println("done");
      } catch (DBusException DBe) {
         test.fail("Failed to add handlers");
      }
      
      System.out.println("Listening for Method Calls");
      testclass tclass = new testclass();
      /** This exports an instance of the test class as the object /Test. */
      conn.exportObject("/Test", tclass);
      
      System.out.println("Sending Signal");
      /** This creates an instance of the Test Signal, with the given object path, signal name and parameters, and broadcasts in on the Bus. */
      conn.sendSignal(new TestSignalInterface.TestSignal("/foo/bar/com/Wibble", "Bar", new UInt32(42)));
      
      System.out.println("Getting our introspection data");
      /** This gets a remote object matching our service name and exported object path. */
      Introspectable intro = (Introspectable) conn.getRemoteObject("foo.bar.Test", "/", Introspectable.class);
      /** Get introspection data */
      String data = intro.Introspect();
      if (null == data || !data.startsWith("<!DOCTYPE"))
         fail("Introspection data invalid");
      System.out.println("Got Introspection Data: \n"+data);
      
      System.out.println("Pinging ourselves");
      /** This gets a remote object matching our service name and exported object path. */
      Peer peer = (Peer) conn.getRemoteObject("foo.bar.Test", "/Test", Peer.class);
      /** Call ping. */
      peer.Ping();
      
      System.out.println("Calling Method0/1");
      /** This gets a remote object matching our service name and exported object path. */
      TestRemoteInterface tri = (TestRemoteInterface) conn.getRemoteObject("foo.bar.Test", "/Test", TestRemoteInterface.class);
      System.out.println("Got Remote Object: "+tri);
      /** Call the remote object and get a response. */
      String rname = tri.getName();
      System.out.println("Got Remote Name: "+rname);
      if (!"This Is A Name!!".equals(rname))
         fail("getName return value incorrect");
        
      /** call something that throws */
      try {
         System.out.println("Throwing stuff");
         tri.throwme();
         test.fail("Method Execution should have failed");
      } catch (TestException Te) {
         System.out.println("Remote Method Failed with: "+Te.getClass().getName()+" "+Te.getMessage());
         if (!Te.getMessage().equals("test"))
            test.fail("Error message was not correct");
      }

      /** Try and call an invalid remote object */
      try {
         System.out.println("Calling Method2");
         tri = (TestRemoteInterface) conn.getRemoteObject("foo.bar.NotATest", "/Moofle", TestRemoteInterface.class);
         System.out.println("Got Remote Name: "+tri.getName());
         test.fail("Method Execution should have failed");
      } catch (ServiceUnknown SU) {
         System.out.println("Remote Method Failed with: "+SU.getMessage());
         if (!SU.getMessage().equals("The name foo.bar.NotATest was not provided by any .service files"))
            test.fail("Error message was not correct");
      }
      
      /** Try and call an invalid remote object */
      try {
         System.out.println("Calling Method3");
         tri = (TestRemoteInterface) conn.getRemoteObject("foo.bar.Test", "/Moofle", TestRemoteInterface.class);
         System.out.println("Got Remote Name: "+tri.getName());
         test.fail("Method Execution should have failed");
      } catch (UnknownObject UO) {
         System.out.println("Remote Method Failed with: "+UO.getClass().getName()+" "+UO.getMessage());
         if (!UO.getMessage().equals("/Moofle is not an object provided by this service."))
            test.fail("Error message was not correct");
      }

      System.out.println("Calling Method4/5/6/7");
      /** This gets a remote object matching our service name and exported object path. */
      TestRemoteInterface2 tri2 = (TestRemoteInterface2) conn.getRemoteObject("foo.bar.Test", "/Test", TestRemoteInterface2.class);
      /** Call the remote object and get a response. */
      
      tri2.dostuff(new Variant(new Boolean(true)));

      short[] shorts = tri2.eatMyShorts();
      if (null == shorts ||
            3 != shorts.length ||
            1 != shorts[0] ||
            2 != shorts[1] ||
            3 != shorts[2])
         test.fail("Bad Shorts");
      
      System.out.println("Sampling Arrays:");
      Integer[] is = tri2.sampleArray(new String[] { "hi", "hello", "hej", "hey", "aloha" }, new Integer[] { new Integer(1), new Integer(5), new Integer(7), new Integer(9) }, new long[] { 2, 6, 8, 12 });
      System.out.println("sampleArray returned an array:");
      for (int i = 0; i < is.length; i++)
         System.out.println("--"+is[i]);
      if (is.length != 5 ||
            is[0].intValue() != -1 ||
            is[1].intValue() != -5 ||
            is[2].intValue() != -7 ||
            is[3].intValue() != -12 ||
            is[4].intValue() != -18)
         fail("sampleArray return value incorrect");

      System.out.print("Sending Array Signal...");
      /** This creates an instance of the Test Signal, with the given object path, signal name and parameters, and broadcasts in on the Bus. */
      conn.sendSignal(new TestSignalInterface.TestArraySignal("/foo/bar/com/Wibble", new String[] { "hi", "hello", "hej", "hey", "aloha" }));
      
      System.out.println("done");

      /** Pause while we wait for the DBus messages to go back and forth. */
      Thread.sleep(1000);
    
      System.out.println("Disconnecting");
      /** Disconnect from the bus. */
      conn.disconnect();
      conn = null;
   } catch (Exception e) {
      e.printStackTrace();
      fail("Unexpected Exception Occurred");
   }}
}
